#include "xkbhandler.h"

#include "kxkbconfig.h"

#include <QDBusConnection>
#include <KConfigGroup>

XkbHandler::XkbHandler(QObject *parent)
    : QObject{parent}
    , m_initialized(false)
    , m_kxkbConfig(KSharedConfig::openConfig(QStringLiteral("kxkbrc"), KConfig::NoGlobals))
    , m_locale1(OrgFreedesktopLocale1Interface(
        QStringLiteral("org.freedesktop.locale1"),
        QStringLiteral("/org/freedesktop/locale1"),
        QDBusConnection::systemBus(),
        this))
    , m_locale1Properties(OrgFreedesktopDBusPropertiesInterface(
        QStringLiteral("org.freedesktop.locale1"),
        QStringLiteral("/org/freedesktop/locale1"),
        QDBusConnection::systemBus(),
        this))
    , m_KDEKeyboardLayouts(OrgKdeKeyboardLayoutsInterface(
          QString(),
          QStringLiteral("/Layouts"),
          QDBusConnection::sessionBus()))
{
    // Sync once at the start
    kxkbChanged();

    connect(&m_KDEKeyboardLayouts,
            &OrgKdeKeyboardLayoutsInterface::layoutListChanged,
            this,
            &XkbHandler::kxkbChanged);

    connect(&m_KDEKeyboardLayouts,
            &OrgKdeKeyboardLayoutsInterface::layoutChanged,
            this,
            &XkbHandler::rotateLayouts);

    connect(&m_locale1Properties,
            &OrgFreedesktopDBusPropertiesInterface::PropertiesChanged,
            this, [this](const QString &interface_name, const QVariantMap &changed, const QStringList &invalidated) {
                if (interface_name == QLatin1StringView(OrgFreedesktopLocale1Interface::staticInterfaceName())
                    && (changed.contains(QStringLiteral("X11Layout")) ||
                        changed.contains(QStringLiteral("X11Model")) ||
                        changed.contains(QStringLiteral("X11Variant")) ||
                        changed.contains(QStringLiteral("X11Options")) ||
                        invalidated.contains(QStringLiteral("X11Layout")) ||
                        invalidated.contains(QStringLiteral("X11Model")) ||
                        invalidated.contains(QStringLiteral("X11Variant")) ||
                        invalidated.contains(QStringLiteral("X11Options")) ))
                    locale1Changed();
            });
}

void XkbHandler::kxkbChanged()
{
    KxkbConfig oldConfig(m_kxkbConfig->group(QStringLiteral("Layout")));
    m_kxkbConfig->reparseConfiguration();
    KxkbConfig newConfig(m_kxkbConfig->group(QStringLiteral("Layout")));

    if (!m_initialized)
        m_initialized = true;
    else if (oldConfig == newConfig)
        return;

    qInfo() << "Write new kxkb to locale1";

    m_locale1.SetX11Keyboard(
        newConfig.LayoutList,
        newConfig.Model,
        newConfig.VariantList,
        newConfig.Options,
        false,
        true);
}

void XkbHandler::locale1Changed()
{
    auto call = m_locale1Properties.GetAll(QLatin1StringView(OrgFreedesktopLocale1Interface::staticInterfaceName()));
    call.waitForFinished();
    if (call.isError())
        return;
    auto x11Properties = call.value();

    m_kxkbConfig->reparseConfiguration();
    auto kConfig = m_kxkbConfig->group(QStringLiteral("Layout"));

    KxkbConfig oldConfig(kConfig);
    KxkbConfig newConfig(x11Properties);

    if (newConfig != oldConfig) {
        qInfo() << "Write new locale1 to kxkb";
        newConfig.write(kConfig);
        m_kxkbConfig->sync();
    }

    // Emit reloadConfig even if the kxkb config hasn't changed.
    // For some reason this helps keep kwin and the plasma applet in sync.
    qInfo() << "Emit reloadConfig";
    QDBusConnection::sessionBus().send(QDBusMessage::createSignal(
        QStringLiteral("/Layouts"),
        QStringLiteral("org.kde.keyboard"),
        QStringLiteral("reloadConfig")));
}

void XkbHandler::rotateLayouts(uint idx)
{
    Q_UNUSED(idx);

    KxkbConfig config(m_kxkbConfig->group(QStringLiteral("Layout")));
    QStringList layouts = config.LayoutList.split(QLatin1Char(','));
    QStringList variants = config.VariantList.split(QLatin1Char(','));
    layouts.append(layouts.takeFirst());
    variants.append(variants.takeFirst());

    config.LayoutList = layouts.join(QLatin1Char(','));
    config.VariantList = variants.join(QLatin1Char(','));

    qInfo() << "Rotate locale1";

    m_locale1.SetX11Keyboard(
        config.LayoutList,
        config.Model,
        config.VariantList,
        config.Options,
        false,
        true);
}
